#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <ftw.h>
#include <ctype.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>

static void do_system(char *p)
{
	printf(">>%s\n", p);
	if(system(p)!=0)
		exit(1);
}

static void do_system_chroot(char *dir, char *p)
{
	int pid = fork();
	int opt;

	switch(pid)
	{
		case -1:
			perror("fork failed");
			exit(1);
		case 0:
			if(chroot(dir) == -1)
			{
				perror("chroot");
				exit(1);
			}
			if(chdir("/") == -1)
			{
				perror("chdir");
				exit(1);
			}
			if(system(p) != 0)
				exit(1);
			exit(0);
		default:
			if(waitpid(pid, &opt, 0) == -1)
				perror("waitpid");
			if(WEXITSTATUS(opt))
				exit(1);
	}
}

static char *archname[64];
static int archnum;

static void do_arch(char *name)
{
	if(archnum == 64)
	{
		fprintf(stderr, "%s: too many architectures.\n", name);
		exit(1);
	}
	archname[archnum] = malloc(strlen(name) + 6);
	if(archname[archnum] == NULL)
	{
		fprintf(stderr, "Out of memory.\n");
		exit(1);
	}
	sprintf(archname[archnum], ".%s.rpm", name);
	archnum++;
}

/*
 *	Check for valid architecture
 */

static int bad_arch(const char *arch)
{
	int i = 0;

	while(archname[i])
	{
		if(strlen(arch) > strlen(archname[i]))
		{
			if(memcmp(arch+strlen(arch)-strlen(archname[i]),
					archname[i], strlen(archname[i]))==0)
				return 0;
		}
		i++;
	}
	return 1;
}

static char *name_seek;
static char *path_found;

static int find_check(const char *file, const struct stat *st, int flag)
{
	const char *fp;
	int n;
	if(flag != FTW_F)
		return 0;
	fp=strrchr(file, '/');
	if(fp)
		fp++;
	else
		fp = file;

	n=strlen(name_seek);

	if(strncmp(fp, name_seek, n) != 0)
		return 0;
	if(fp[n] != '-' && fp[n])
		return 0;
	if(!isdigit(fp[n+1]))
		return 0;
	if(bad_arch(file))
		return 0;
	path_found = strdup(file);
	if(path_found == NULL)
	{
		fprintf(stderr, "Out of memory.\n");
		exit(1);
	}
	return 1;
}

static char *find_rpm(char *name)
{
	name_seek = name;
	path_found = NULL;
	if(ftw(".", find_check, 8) == 0 || path_found == NULL)
	{
		fprintf(stderr, "RPM %s not found.\n", name);
		exit(1);
	}
	return path_found;
}

static int rpmset(char *rootdir, char *rpm0, char *rest, int retry)
{
	char *pathbuf[70];
	char **paths = &pathbuf[0];
	char *names[64];
	int ct = 1;
	int pid;
	int opt;

	*paths++ = "rpm";
	*paths++ = "--root";
	*paths++ = rootdir;
	*paths++ = "-ivh";
	if(retry)
		*paths++ = "--nodeps";

	names[0] = rpm0;
	while(ct < 64)
	{
		names[ct] = strtok_r(rest, " \t", &rest);
		if(names[ct] == NULL)
			break;
		ct++;
	}
	if(ct == 64)
	{
		fprintf(stderr, "Too many dependant rpms.\n");
		exit(1);
	}
	for(ct = 0; names[ct]; ct++)
		paths[ct] = find_rpm(names[ct]);
	paths[ct] = NULL;

	switch(pid = fork())
	{
		case -1:
			perror("fork failed");
			exit(1);
		case 0:
			execvp("rpm", pathbuf);
			perror("rpm");
			exit(1);
		default:
			if(waitpid(pid, &opt, 0)==-1)
				perror("waitpid");
	}
	for(ct = 0; names[ct]; ct++)
		free(paths[ct]);
	return WEXITSTATUS(opt);
}

	
static void do_install(char *root)
{
	char buf[1024];
	while(fgets(buf, 1024, stdin))
	{
		char *h;
		char *t;
		int f;
	
		t = strchr(buf, '#');
		if(t)
			*t = 0;
		t = strchr(buf, '\n');
		if(t)
			*t=0;
	
		h=strtok_r(buf, " \t", &t);
		if(h==NULL)
			continue;
		if(strcmp(h, "%endinstall") == 0)
			break;
		if(strcmp(h, "%disc") == 0)
		{
			printf("Change to disc %s.\n", t);
			continue;
		}
	
		f = 0;
		if(*h=='%')
		{
			if(strcmp(h+1, "nodep")==0)
				f = 1;
			else
				printf("Conditional class (%s).\n", h+1);
			h=strtok_r(t, " \t", &t);
			if(h==NULL)
				continue;
		}
		if(rpmset(root, h, t, f)!=0)
		{
			exit(1);
			if(rpmset(root, h, t, 1)!=0)
			{
				fprintf(stderr, "RPM install failed.\n");
				exit(1);
			}
		}
	}
}

static void environment(char *e, char *v)
{
	char buf[1024];
	char *x;
	snprintf(buf, 1024, "%s=%s", e, v);
	x=strdup(buf);
	if(x!=NULL && putenv(x)!=-1)
		return;
	fprintf(stderr, "Setting enviroment %s=%s failed.\n", e, v);
	exit(1);
}

int main(int argc, char *argv[])
{
	char buf[1024];

	signal(SIGCHLD, SIG_DFL);

	if(argc!=6)
	{
		fprintf(stderr,"%s rpmdir rootdir hostname nisname nisserver.\n", argv[0]);
		exit(1);
	}
	if(chdir(argv[1]) == -1)
	{
		perror(argv[1]);
		exit(1);
	}
	if(mkdir(argv[2], 0700) == -1)
	{
		perror(argv[2]);
		exit(1);
	}
	environment("RPMDIR", argv[1]);
	environment("INSTALL", argv[2]);
	environment("HOSTNAME", argv[3]);
	environment("NISDOMAIN", argv[4]);
	environment("NISSERVER", argv[5]);

	while(fgets(buf,1024, stdin))
	{
		char *h;
		char *t = strchr(buf, '#');
		if(t)
			*t = 0;
		t = strchr(buf, '\n');
		if(t)
			*t=0;
		h=strtok_r(buf, " \t", &t);
		if(h == NULL)
			continue;
		if(strcmp(h, "%end") == 0)
			break;
		else if(strcmp(h, "%arch") == 0)
			do_arch(t);
		else if(strcmp(h, "%run") == 0)
			do_system(t);
		else if(strcmp(h, "%runin") == 0)
			do_system_chroot(argv[2], t);
		else if(strcmp(h, "%install") == 0)
			do_install(argv[2]);
		else fprintf(stderr, "Unknown command - %s.\n", h);
	}
	return 0;
}
